(*  Title:      /ParRed.thy
    ID:         $Id: ParRed.thy,v 1.2 2007/04/19 17:42:02 hsudhof Exp $
    Author:     Ludovic Henrio and Florian Kammuller
    Copyright   2006

confluence of beta for sgma calculus, based on the equally named file in HOL/Lambda

Defines a parallel reduction of sigma terms and proves confluence
*)

header {* Parallel reduction *}

theory ParRed imports FinASP Commutation begin


subsection {* Parallel reduction *}

consts
  par_beta :: "(dB \<times> dB) set"

syntax
  par_beta :: "[dB, dB] => bool"  (infixl "=>" 50)
translations
  "s => t" == "(s, t) \<in> par_beta"

inductive par_beta
  intros
    var [simp, intro!]: "Var n => Var n"
    obj [simp, intro!]: "\<lbrakk> dom s = dom s'; \<forall>  l \<in> dom s . the (s l) => the (s' l) \<rbrakk> \<Longrightarrow> Obj s => Obj s'"
    upd [simp, intro!]: "\<lbrakk> s => s'; t => t' \<rbrakk> \<Longrightarrow> Upd s l t => Upd s' l t'"
    upd' [simp, intro!]:"\<lbrakk> Obj s => Obj s'; t => t'; l \<in> dom s \<rbrakk> 
                         \<Longrightarrow> (Upd (Obj s) l t)  => (Obj (s'(l \<mapsto> t')))"
    sel [simp, intro!]: "s => t \<Longrightarrow> Call s l => Call t l"
    beta [simp, intro!]: "\<lbrakk> Obj f => Obj f'; l \<in> dom f' \<rbrakk>  \<Longrightarrow> Call (Obj f) l => (the (f' l))[(Obj f')/0]"


(* These are rule formats for inversions rules (try ML "thms \"par_preta_cases\"") *)
inductive_cases par_beta_cases [elim!]:
  "Var n => t"
  "Obj s => t"
  "Call s l => t"
  "Upd s l t => u"


subsection {* Inclusions *}

text {* @{text "beta \<subseteq> par_beta \<subseteq> beta^*"} \medskip *}

lemma par_beta_varL [simp]:
    "(Var n => t) = (t = Var n)"
  apply blast
  done


lemma par_beta_obj_subst': "\<lbrakk> l \<in> dom  f; s => t; Obj f => Obj f' \<rbrakk> 
                         \<Longrightarrow> Obj (f (l \<mapsto> s)) => Obj (f'(l \<mapsto> t))";

apply (rule obj);
apply fastsimp;
apply (rule ballI);

apply (case_tac "la = l");
apply fastsimp;

apply(simp)
apply(blast)
done;


lemma par_beta_obj_subst: "\<lbrakk> s => t; Obj f => Obj f' \<rbrakk> 
                         \<Longrightarrow> Obj (f (l \<mapsto> s)) => Obj (f'(l \<mapsto> t))";

apply (case_tac "l \<in> dom f");
apply (simp add: par_beta_obj_subst');
apply force;
done;


lemma par_beta_refl_prep: "t => t & Obj l => Obj l";
apply (rule db_induct2);
apply simp_all;

apply(rule par_beta.intros)

apply(simp);
apply(rule ballI)
apply(case_tac "la=l")
apply(simp)
apply(simp)
apply force;

done;

lemma par_beta_refl [simp]: "t => t"  (* par_beta_refl [intro!] causes search to blow up *)
(* apply (simp add: par_beta_refl_prep); *)
apply (rule dB_induct);
apply simp_all; 
apply (subgoal_tac "t => t & Obj l => Obj l");
apply (erule conjE);
apply assumption;
apply (rule par_beta_refl_prep);
done;


lemma beta_subset_par_beta: "beta <= par_beta"
  apply (rule subsetI)
  apply clarify
  apply (erule beta.induct)
  defer;
  defer;
     apply (blast intro!: par_beta_refl)+
     apply (rule obj);
  apply fastsimp+;

done;

lemma par_beta_subset_beta: "par_beta <= beta^*"
  apply (rule subsetI)
  apply clarify
  apply (erule par_beta.induct)
     apply blast
apply (erule rtrancl_beta_obj_n);
apply simp;
    defer;
    defer;
apply blast;
    apply (rule rtrancl_into_rtrancl);
    defer;
    apply (rule FinASP.beta);
    apply assumption;
    apply (rule rtrancl_trans);
    apply (erule rtrancl_beta_updL);
    apply (erule rtrancl_beta_updR);



defer
    apply (rule rtrancl_beta_Obj);
    apply assumption;


    apply (rule rtrancl_into_rtrancl);
    apply (rule rtrancl_trans);
    apply (erule rtrancl_beta_updL);
    apply (erule rtrancl_beta_updR);
    apply (rule beta.intros);
apply(force)
done;


(* lemmata for inversion of par_beta *)
lemma Obj_par_red: "\<lbrakk> Obj s => z \<rbrakk> \<Longrightarrow> ? lz. dom s = dom lz & z = Obj lz";

(* ML "thms \"par_beta_cases\""; *)
apply (erule par_beta_cases);
apply (rule_tac x = s' in exI);
apply simp;
done;


lemma Upd_par_red: "\<lbrakk> Upd s l t => z \<rbrakk> \<Longrightarrow> 
        (? s' t'. s => s' & t => t' & z = Upd s' l t') | 
        (? f f' t'. s = Obj f & Obj f => Obj f' & t => t' & z = Obj (f' (l \<mapsto> t')))";

apply (erule par_beta_cases);
apply (rule disjI1);
apply (rule_tac x = s' in exI);
apply (rule_tac x = t' in  exI);
apply simp;
apply (rule disjI2);
apply (rule_tac x = sa in exI);
apply (rule_tac x = s' in exI);
apply (rule_tac x = t' in exI);
apply simp;
done;

lemma Call_par_red: "\<lbrakk> Call s l => z \<rbrakk> \<Longrightarrow> (? t. s => t & z = Call t l)
         | (? f f'. Obj f => Obj f' & l \<in> dom f' & s = Obj f & z = (the (f' l))[(Obj f')/0])";

apply (erule par_beta_cases);
apply (rule disjI1);
apply (rule_tac x = t in exI);
apply simp;
apply (rule disjI2);
apply (rule_tac x = f in  exI);
apply (rule_tac x = f' in  exI);
apply simp;
done;




subsection {* Misc properties of par-beta *}

lemma list_decon: "l ~= [] \<Longrightarrow> l = hd l # tl l";
apply auto;
done

(*
lemma lift_obj_lem0: "Obj (x # l) => t \<Longrightarrow> ? x' l'. length l = length l' & t = Obj (x' # l')";
apply (frule Obj_par_red);
apply (erule exE);
apply (rule_tac x = "hd lz" in exI);
apply (rule_tac x = "tl lz" in exI);
apply (simp add: list_decon);
apply (rule conjI);
apply arith;
apply (rule list_decon);
apply fastsimp;
done;


lemma lift_obj_lem00: "Obj (x # l) => Obj (x' # l') \<Longrightarrow> x => x'";
apply fastsimp;
done;


lemma lift_obj_lem000: "Obj (x # l) => Obj (x' # l') \<Longrightarrow> Obj l => Obj l'";
apply fastsimp;
done;


lemma pred_hd_nth: "0 < n \<longrightarrow> (x # l) ! n = l ! (n - 1)"
apply (induct_tac n);
apply auto;
done;

lemma pred_hd_nth': "0 < n \<Longrightarrow> (x # l) ! n = l ! (n - 1)"
apply (simp add: pred_hd_nth);
done;


lemma obj_suc_lem[rule_format]: "\<lbrakk> Obj l => Obj l'; x => x'\<rbrakk> \<Longrightarrow>(Obj (x # l)) => (Obj (x' # l'))";
apply (rule obj);
apply fastsimp;
apply auto;

apply (case_tac "la = 0");
apply simp;
apply (subgoal_tac "0 < la");
apply (subgoal_tac "l  ! (la - 1) => l' ! (la - 1)");
apply (simp add: pred_hd_nth');
apply (frule_tac x = "la - 1" in spec);
apply (erule impE);
apply arith;
apply assumption;
apply arith;
done;

*)

lemma lift_obj_lem0: "Obj (f) => t \<Longrightarrow> ? f'. dom f = dom f' & t = Obj (f')";
apply (frule Obj_par_red);
apply (erule exE);

apply (rule_tac x=lz in exI);
apply(assumption)

done;


lemma lift_obj_lem00: "Obj (f) => Obj (f') \<Longrightarrow>  dom f = dom f'";
apply(auto)


done;


lemma dom_liftoption_lem: " !!f .  dom f = dom (\<lambda> l. lift_option n (f l) )";




apply auto;
apply(case_tac "x\<in>  dom f")
apply(force)
apply(force)

done;


lemma insert_dom_eq : "\<lbrakk>dom f = dom f'\<rbrakk> \<Longrightarrow> dom (f(l |-> x)) = dom (f'(l \<mapsto> x'))"
apply(auto)
done

lemma insert_dom_less_eq : "!! y y' x .  \<lbrakk>x \<notin> dom f; x \<notin> dom f';  dom (f(x \<mapsto> y)) = dom (f'(x \<mapsto> y'))\<rbrakk> \<Longrightarrow> dom f = dom f'"
apply(auto)
apply(case_tac "xa = x")
apply(simp)
apply(blast)
apply(case_tac "xa = x")
apply(simp)
apply(blast)
done


lemma par_beta_less : "!! d1 t1 . \<lbrakk> Obj (f(l\<mapsto>d1)) => Obj(t(l |-> t1)) ; l \<notin> dom f; l \<notin> dom t \<rbrakk> \<Longrightarrow> Obj f =>  Obj t"
apply(subgoal_tac "dom f = dom t")
prefer 2
apply(subgoal_tac " dom (f(l |-> d1)) = dom  ((t(l |-> t1)))")
prefer 2
apply(rule lift_obj_lem00)
apply(assumption)
apply(subgoal_tac " dom (f(l \<mapsto> d1)) = dom f \<union> {l}")
apply(auto)
apply(erule_tac x=la in ballE)
apply(case_tac  "la=l")
apply(auto)
done

lemma par_beta_less' : "!! d1 t1 . \<lbrakk> \<forall> x \<in> dom f.  the ((f(l\<mapsto>d1)) x) => the ((t(l |-> t1)) x) ; l \<notin> dom f; l \<notin> dom t \<rbrakk> \<Longrightarrow> \<forall> x \<in> dom f .  the (f x) =>  the (t x)"
apply(rule ballI)
apply(erule_tac x=x in ballE)
apply(case_tac "x=l")
apply(auto)
done


lemma par_beta_one : "!! d1 t1 . \<lbrakk> Obj (f(l\<mapsto>d1)) => Obj(t(l |-> t1)) ; l \<notin> dom f; l \<notin> dom t \<rbrakk> \<Longrightarrow>  d1 =>  t1"
apply(subgoal_tac "dom f = dom t")
prefer 2
apply(subgoal_tac " dom (f(l |-> d1)) = dom  ((t(l |-> t1)))")
prefer 2
apply(rule lift_obj_lem00)
apply(assumption)
apply(subgoal_tac " dom (f(l \<mapsto> d1)) = dom f \<union> {l}")
apply(auto)

done


lemma lift_obj_lem: "\<lbrakk> lift (Obj f) n => lift (Obj f') n; lift x (Suc n) => lift x' (Suc n)\<rbrakk>
                     \<Longrightarrow> lift (Obj (f(l \<mapsto> x))) n => lift (Obj (f'(l |-> x'))) n";


apply (subgoal_tac "  (lift (Obj (f(l \<mapsto> x))) n => lift (Obj (f'(l |-> x'))) n ) = (Obj (\<lambda>la. lift_option (Suc n) ( (f(l \<mapsto> x)) la)) =>
       Obj (\<lambda>la. lift_option (Suc n) ((f'(l \<mapsto> x')) la))) ")
prefer 2
apply(simp)
apply(rotate_tac -1)
apply(erule ssubst)
(*
simp+auto + force would be enough, but we do it by hand*)
apply(rule obj)
apply(subgoal_tac "dom (\<lambda>la. lift_option (Suc n) ((f(l \<mapsto> x))la) ) = dom (f(l \<mapsto> x))")
prefer 2;
apply(rule sym)
apply(rule dom_liftoption_lem)
apply(subgoal_tac "dom (\<lambda>la. lift_option (Suc n) ((f'(l \<mapsto> x'))la) ) = dom (f'(l \<mapsto> x'))")
prefer 2;
apply(rule sym)
apply(rule dom_liftoption_lem)
apply(rotate_tac -1)
apply(erule ssubst)
apply(rotate_tac -1)
apply(erule ssubst)

apply(subgoal_tac " dom f = dom f'")
prefer 2;
apply(simp)
apply(subgoal_tac " dom (\<lambda>l. lift_option (Suc n) (f l))  = dom (\<lambda>l. lift_option (Suc n) (f' l))")
prefer 2
apply(rule lift_obj_lem00)
apply(assumption)
apply(subgoal_tac " dom f'  = dom (\<lambda>l. lift_option (Suc n) (f' l))")
prefer 2
apply(rule dom_liftoption_lem)
apply(subgoal_tac " dom f  = dom (\<lambda>l. lift_option (Suc n) (f l))")
prefer 2
apply(rule dom_liftoption_lem)
apply(rotate_tac -1)
apply(erule ssubst)
apply(rotate_tac -1)
apply(erule ssubst)
apply(assumption)
apply(rule insert_dom_eq)
apply(assumption)

apply(rule ballI)
apply(case_tac "la = l")
apply(simp)
apply(subgoal_tac "((f(l \<mapsto> x)) la) = f la" )
prefer 2
apply(simp)
apply(subgoal_tac "((f'(l \<mapsto> x')) la) = f' la" )
prefer 2
apply(simp)
apply(rotate_tac -1)
apply(erule ssubst)
apply(rotate_tac -1)
apply(erule ssubst)
apply(force)
done



lemma dom_substoption_lem: "dom (\<lambda> l . subst_option n s  (f l) ) = dom f";
apply(auto)
apply(case_tac "x \<in>  dom f")
apply(force)
apply(force)

done;


lemma olift_lift: "! n'. n' \<in> dom f \<longrightarrow> (the ((\<lambda> l . lift_option n (f l)) n')) = (lift (the (f n')) n)";
apply(auto)
done





lemma olift_upd: "! n'. ((\<lambda> l . lift_option (Suc n)  ((s(n'|-> t)) l ))  =  ((\<lambda> l . lift_option (Suc n)(s l))( n' \<mapsto> (lift t (Suc n)))))";
apply(rule allI)
apply(rule ext)
apply(case_tac "n'=l")
apply(simp)
apply(simp)
done


lemma  par_beta_beta : "!!a b. \<lbrakk> a -> b\<rbrakk> \<Longrightarrow> a => b"
apply(subgoal_tac "beta \<subseteq> par_beta")
apply(rule set_mp)
apply(simp)
apply(assumption)
apply(rule beta_subset_par_beta)
done

lemma insert_one : "\<lbrakk>l\<in> dom t\<rbrakk> \<Longrightarrow> \<exists> t' . l \<notin> dom t' \<and> t = (t'(l \<mapsto> (the (t l))))"
apply(rule_tac x="\<lambda> la . if la = l then None else t la" in exI )
apply(rule conjI)
apply(force)

apply(rule ext)
apply(force)

done

lemma obj_par_beta : "\<lbrakk>Obj f => Obj f'\<rbrakk> \<Longrightarrow> \<forall>l\<in>  dom f . the (f l) => the (f' l)"
apply auto
done


lemma obj_par_beta' : "\<lbrakk> \<forall>l\<in>  dom f . the (f l) => the (f' l); dom f = dom f'\<rbrakk> \<Longrightarrow>  Obj f => Obj f'"
apply auto
done


lemma par_beta_lift_prep [rule_format, simp]:
    "(\<forall>t' n. t => t' --> lift t n => lift t' n)
    &(\<forall>t' n. Obj f => Obj t' --> lift (Obj f) n => lift (Obj t') n)";
apply (rule db_induct2);
apply fastsimp;
apply fastsimp;
apply(rule allI)
apply(rule allI)
apply(rule impI)
apply(erule par_beta.elims)
apply(simp)
apply(simp)
apply(simp)
apply(simp)
apply(simp)

apply(subgoal_tac "dom f = dom f'")
prefer 2
apply(rule lift_obj_lem00);
apply(assumption)



apply(subgoal_tac "t'= the (f' l)[Obj f'/0]")
prefer 2
apply(simp)
apply(rotate_tac -1)
apply(erule ssubst)
apply(subgoal_tac " lift (Call d l) n =  lift (Call (Obj f) l) n ")
prefer 2
apply(simp)
apply(rotate_tac -1)
apply(erule ssubst)
apply(erule_tac x="Obj f'" in allE)
apply(erule_tac x=n in allE)
apply(erule impE)

apply(simp)
apply(subgoal_tac "l \<in> dom f")
prefer 2
apply(simp)
apply(subgoal_tac "la \<in> dom f")
prefer 2
apply(simp)

apply(subgoal_tac " lift (Call (Obj f) l) n =>  lift (Call (Obj f') l) n ")
prefer 2
apply(simp)
apply(subgoal_tac " lift (Call (Obj f') l) n => lift (the (f' l)[Obj f'/0]) n ")

prefer 2
apply(subgoal_tac "Call (Obj f') l => the (f' l)[Obj f'/0]")
prefer 2
apply(simp)

apply(simp)


apply(subgoal_tac "(Call (Obj (\<lambda>l. lift_option (Suc n) (f' l))) la) ->
         (the ((\<lambda>l. lift_option (Suc n) (f' l)) la)  [(Obj (\<lambda>l. lift_option (Suc n) (f' l)))/0])")
prefer 2

apply(rule beta.intros)
apply(subgoal_tac "dom f' = dom (\<lambda>l. lift_option (Suc n) (f' l))")
prefer 2

apply(rule dom_liftoption_lem)
apply(rotate_tac -1)
apply(erule subst)
apply(assumption)
apply(subgoal_tac "the (lift_option (Suc n)(f' la)) =  lift  (the (f' la))  (Suc n)")
prefer 2
apply(force)
apply(rotate_tac -1)

apply(erule subst)
apply(simp add: par_beta_beta)

apply(subgoal_tac "(lift (Call (Obj f) l) n) = (Call (Obj (\<lambda>la\<Colon>Label. lift_option (Suc n) (f la))) l)")
prefer 2
apply(simp)


apply(simp)
apply(subgoal_tac " lift (the (f' la))(Suc n) = (the( ( \<lambda>l\<Colon>Label. lift_option (Suc n) (f' l)) la))") 
prefer 2
apply(force)

apply(rotate_tac -1)
apply(erule ssubst)
apply(rule par_beta.intros)
apply(simp)

apply(subgoal_tac "dom f' = dom  (\<lambda>l\<Colon>Label. lift_option (Suc n) (f' l))")
prefer 2
apply(rule dom_liftoption_lem)
apply(rotate_tac -1)
apply(erule subst)
apply(assumption)
(* call done *)


apply(rule allI)
apply(rule allI)
apply(rule impI)
apply(erule par_beta.elims)
apply(simp)
apply(simp)
apply(simp)
apply(subgoal_tac "lift (Upd (Obj s) la t) n => lift (Obj (s'(la \<mapsto> t'a))) n" )
apply(simp)
apply(subgoal_tac "lift (Upd (Obj s) la t) n = Upd (lift (Obj s) n) la (lift t (Suc n))" )
prefer 2
apply(simp)
apply(subgoal_tac "lift (Obj (s'(la \<mapsto> t'a))) n = Obj (\<lambda> l . lift_option (Suc n) ((s'(la \<mapsto> t'a)) l)) ")
prefer 2

apply(simp)

apply(rotate_tac -1)
apply(erule ssubst)
apply(rotate_tac -1)
apply(erule ssubst)
apply(rotate_tac 1)
apply(erule_tac x="t'a" in allE)
apply(erule_tac x="Suc n" in allE)
apply(erule impE)
apply(simp)
apply(subgoal_tac " lift d2 (Suc n) =  (lift t (Suc n))" )
prefer 2
apply(simp)
apply(rotate_tac -1)
apply(erule ssubst)
apply(subst olift_upd)
apply(simp)
apply(erule_tac x="Obj s'" in allE)
apply(erule impE)
apply(assumption)
apply(erule_tac x="n" in allE)
apply( rule upd')
apply(simp)
apply(assumption)
apply(subgoal_tac  "dom s = dom (\<lambda>l\<Colon>Label. lift_option (Suc n) (s l))")
apply(simp)
apply(rule dom_liftoption_lem)
apply(force)

apply(force)
(* update done*)
apply(rule allI)
apply(rule allI)
apply(rule impI)
apply(simp)
apply(subgoal_tac "dom empty = dom t'")
apply(subgoal_tac " t' = empty")
apply(force)
apply(simp)
apply(rule empty_dom)
apply(assumption)
apply(force)

(* we are here *)
apply(rule allI)
apply(rule allI)
apply(rule impI)
apply(subgoal_tac "l \<in> dom t'")
prefer 2

apply(force)
apply(subgoal_tac "dom (f(l \<mapsto> d1)) = dom t'")


prefer 2

apply(rule lift_obj_lem00)
apply(assumption)


apply(frule insert_one)
apply(erule exE)
apply(erule conjE)


apply(subgoal_tac " Obj f => Obj t'a")
prefer 2
apply(subgoal_tac "Obj (f(l \<mapsto> d1)) => Obj (t'a(l \<mapsto> the (t' l) ))")
prefer 2

apply(rotate_tac -1)
apply(erule subst)
apply(assumption)

apply(rule_tac f=f and l=l and t="t'a"  in  par_beta_less)
apply(blast)
apply(assumption)
apply(simp)

apply(subgoal_tac "Obj t' = Obj (t'a(l \<mapsto> the (t' l)))")
apply(rotate_tac -1)
apply(erule ssubst)
apply(rule lift_obj_lem)
apply(simp)
apply(erule_tac x="the (t' l)" in allE)
apply(erule_tac x="Suc n" in allE)
apply(erule impE)
thm obj_par_beta
apply(frule_tac f="f(l \<mapsto> d1)" and f'=t' in  obj_par_beta)
apply(erule_tac x=l in ballE)
apply(simp)
apply(simp)
apply(assumption)
apply(simp)


done;
(*
lemma nth_pred[rule_format]: "s \<noteq> [] \<longrightarrow> s ! (l - Suc 0) = tl s ! l";
apply (induct_tac s);
apply simp;
apply simp;
thm nth_Cons_Suc;
*)

lemma osubst_lift[rule_format]: "\<forall>  n' \<in> dom  f.
             (the (f n'))[lift s 0/Suc n] =  the ((subst_option  (Suc n)  (lift s 0)  (f n')))";
apply(auto)
done





lemma par_beta_subst_prep [rule_format]:
    "(\<forall>s s' t' n. s => s' --> t => t' --> t[s/n] => t'[s'/n])
    &(\<forall>s s' t' n. s => s' --> (Obj f) => t' --> Obj f[s/n] => t'[s'/n])";
apply (rule db_induct2);

apply (simp add: subst_Var);
apply(simp)

apply (intro strip);
apply (erule par_beta_cases);
apply simp;


apply(subgoal_tac "Call (Obj f') l[s/n] = Call ((Obj f')[s/n]) l")
prefer 2
apply(simp)
apply(subgoal_tac "Call d  l[s/n] = Call ((Obj f)[s/n]) l")
prefer 2
apply (simp)
apply(rotate_tac -1)
apply(erule ssubst)
apply(subgoal_tac " Call (Obj f[s/n]) l =>  Call (Obj f'[s'/n]) l")
prefer 2
apply(erule_tac x=s in allE)
apply(erule_tac x=s' in allE)
apply(erule_tac x="Obj f'" in allE)
apply(erule_tac x="n" in allE)
apply(erule impE)

apply(assumption)
apply(erule impE)
apply(simp)


apply(hypsubst)
apply(simp)
apply(subgoal_tac "Obj f [s/n] = Obj (\<lambda> l . subst_option (Suc n) (lift s 0) (f l))")
prefer 2
apply(simp)
apply(subgoal_tac "Obj f'[s'/n]= Obj((\<lambda> la . subst_option (Suc n) (lift s' 0) (f' la)) )")
prefer 2
apply(simp)
apply(subgoal_tac " Obj (\<lambda> l . subst_option (Suc n) (lift s 0) (f l)) => Obj((\<lambda> la . subst_option (Suc n) (lift s' 0) (f' la)) )")
prefer 2
apply(rotate_tac -1)
apply(erule subst)
apply(rotate_tac -1)
apply(erule subst)
apply(erule_tac x=s in allE)
apply(erule_tac x=s' in allE)
apply(erule_tac x="Obj f'" in allE)
apply(erule_tac x="n" in allE)
apply(erule impE)
apply(assumption)

apply(erule impE)
apply(simp)
apply(simp)

apply(subgoal_tac "the (f' l)[Obj f'/0][s'/n] = the ((\<lambda>la. subst_option (Suc n) (lift s' 0) (f' la)) l)[(Obj f'[s'/n])/0] ")
apply(subgoal_tac " t'[s'/n] =  the (subst_option (Suc n) (lift s' 0) (f' l))[Obj f'[s'/n]/0]")
prefer 2
apply(simp)
apply(rotate_tac -1)
apply(erule ssubst)
apply (subgoal_tac "the (subst_option (Suc n) (lift s' 0) (f' l)) =  the (f' l)[lift s' 0/Suc n]" )
apply(subgoal_tac "  the (subst_option (Suc n) (lift s' 0) (f' l))[Obj f'[s'/n]/0] =   the (f' l)[Obj f'/0][s'/n]")
prefer 2
apply(simp)

apply(rotate_tac -6)
apply(erule ssubst)
apply(subgoal_tac "l \<in>  dom f")
thm beta
apply(subgoal_tac "Obj f'[s'/n]  = Obj (\<lambda> l .  subst_option (Suc n) (lift s' 0) (f' l))")
prefer 2
apply(force)
apply(rotate_tac -1)
apply(erule ssubst)
apply(subgoal_tac "l \<in> dom (\<lambda> l .  subst_option (Suc n) (lift s' 0) (f' l))")
apply(rule_tac  f= "\<lambda>l. subst_option (Suc n) (lift s 0) (f l)" and f'="\<lambda> l .  subst_option (Suc n) (lift s' 0) (f' l)" and l=l in par_beta.beta)
apply(assumption)
apply(assumption)
apply(simp add: dom_substoption_lem)
apply(subgoal_tac "dom f = dom f'")
apply(simp)
apply(rule lift_obj_lem00)
apply(assumption)
apply(force)
apply(subgoal_tac "the (subst_option (Suc n) (lift s' 0) (f' l)) = the (f' l)[(lift s' 0)/Suc n]")
apply(rotate_tac -1)
apply(erule ssubst)
apply(subst subst_subst)
apply(arith)
apply(rule refl)
apply(force)
(*CALL done*)



apply(rule allI)+
apply(rule impI)
apply(rule impI)
thm par_beta.cases
apply(simp)
apply(erule_tac  par_beta_cases)
apply(simp)


apply(subgoal_tac  "Obj (s'a(l \<mapsto> t'a))[s'/n] =  Obj (\<lambda>  la. (subst_option (n+1)(lift s' 0)((s'a(l \<mapsto> t'a)) la)))")
prefer 2
apply(simp)
apply(subgoal_tac "Obj (\<lambda>  la. (subst_option (n+1)(lift s' 0)((s'a(l \<mapsto> t'a)) la))) 
                 = Obj ((\<lambda>  la. (subst_option (n+1)(lift s' 0)((s'a) la))) (l \<mapsto> t'a[(lift s' 0)/n + 1] ))")
prefer 2
apply(simp)
apply(rule ext)
apply(simp)
apply(subgoal_tac  "Obj sa[s/n] =  Obj (\<lambda>  la. (subst_option (n+1)(lift s 0)(sa la)))")
prefer 2
apply(simp)
apply(subgoal_tac  "Obj s'a[s'/n] =  Obj (\<lambda>  la. (subst_option (n+1)(lift s' 0)(s'a la)))")
prefer 2
apply(simp)

apply(erule_tac x="s" in allE)
apply(erule_tac x="lift s 0" in allE)
apply(erule_tac x="s'" in allE)
apply(erule_tac x="lift s' 0" in allE)
apply(erule impE)
apply(assumption)
apply(erule impE)
apply(simp)
apply(erule_tac x="Obj s'a" in allE)
apply(erule_tac x="t'a" in allE)
apply(erule impE)
apply(simp)
apply(erule impE)
apply(simp)
apply(erule_tac x="n" in allE)
apply(erule_tac x="n+1" in allE)
apply(force)
(* update done *)

apply(rule allI)
apply(rule allI)
apply(rule allI)
apply(rule allI)
apply(rule impI)
apply(rule impI)
apply(subgoal_tac "\<exists> t . t' = Obj t")
prefer 2
apply(force)
apply(erule exE)
apply(subgoal_tac "dom empty = dom t")
apply(subgoal_tac " t = empty")
apply(simp)

apply(simp add: empty_dom)
apply(rule  lift_obj_lem00)
apply(simp)

apply(rule allI)+

apply(rule impI)+
apply(subgoal_tac "\<exists> t . t' = Obj t")
prefer 2
apply(force)
apply(erule exE)
apply(subgoal_tac "l \<in> dom t")
prefer 2

apply(subgoal_tac "dom (f(l \<mapsto> d1)) = dom t")
apply(subgoal_tac "l \<in> dom (f(l \<mapsto> d1))")
apply(simp)
apply(force)


apply(rule lift_obj_lem00)
apply(rotate_tac -1)
apply(erule subst)
apply(assumption)


apply(frule insert_one)
apply(erule exE)

apply(erule conjE)

apply(subgoal_tac  "t'[s'/n] = Obj ( t'a(l \<mapsto> the (t l)))[s'/n]")
apply(rotate_tac -1)
apply(erule ssubst)
prefer 2
apply(simp)
apply(subgoal_tac "Obj (f(l \<mapsto> d1))[s/n] = Obj((\<lambda> la . subst_option (n + 1) (lift s 0) (f la))(l \<mapsto> d1[(lift s 0)/n + 1]) ) ")
apply(subgoal_tac " Obj (t'a(l \<mapsto> the (t l)))[s'/n]  = Obj((\<lambda> la . subst_option (n + 1) (lift s' 0) (t'a la))(l \<mapsto>  the (t l) [(lift s' 0)/n + 1]) ) ")
apply(subgoal_tac "Obj f => Obj t'a") 
prefer 2
thm par_beta_less
apply(subgoal_tac "Obj  (f(l \<mapsto> d1)) => Obj  (t'a(l \<mapsto> the (t l)))") 
prefer 2
apply(subgoal_tac "Obj  (f(l \<mapsto> d1)) => Obj  t") 
apply(force)
apply(blast)
thm par_beta_less
apply(frule_tac f=f and l=l in par_beta_less)
apply(assumption)
apply(assumption)
apply(assumption)

(*I am here*)
apply (erule_tac x="lift s 0" in allE )
apply (erule_tac x="s" in allE )
apply (erule_tac x="lift s' 0" in allE )
apply (erule_tac x="s'" in allE )
apply (erule_tac x="the (t l)" in allE )
apply (erule_tac x="Obj t'a" in allE )
apply (erule_tac x="n + 1" in allE )
apply (erule_tac x="n" in allE )
apply(erule impE)
apply(simp)
apply(erule impE)
apply(simp)
apply(erule impE)


apply(subgoal_tac " d1 = the ( (f(l \<mapsto> d1)) l)")
prefer 2
apply(simp)
apply(rotate_tac -1)
apply (erule ssubst)
thm obj_par_beta
apply(subgoal_tac " Obj (f(l \<mapsto> d1)) => Obj t")
apply(frule_tac f="f(l \<mapsto> d1)"  and f'= t in  obj_par_beta)
apply(erule_tac x=l in ballE)
apply(assumption)

apply(simp)
apply(blast)
apply(erule impE)
apply(simp)
apply(simp)
apply(rule_tac  obj_par_beta')
apply(rule ballI)
apply(case_tac "l=la")
apply(simp)
apply(simp)
apply(rule impI)
apply(erule disjE)
apply(frule_tac f="\<lambda>l. subst_option (Suc n) (lift s 0) (f l)" and f'="\<lambda>l. subst_option (Suc n) (lift s' 0) (t'a l)" in  obj_par_beta)
apply(erule_tac x=la in ballE)

apply(simp)
apply(simp)

apply(frule_tac f="\<lambda>l. subst_option (Suc n) (lift s 0) (f l)" and f'="\<lambda>l. subst_option (Suc n) (lift s' 0) (t'a l)" in  obj_par_beta)
apply(erule_tac x=la in ballE)
apply(assumption)

apply(subgoal_tac " dom ((\<lambda>la. subst_option (Suc n) (lift s 0) (f la))(l \<mapsto> d1[lift s 0/Suc n])) = dom ((f)(l \<mapsto> d1[lift s 0/Suc n])) ")
apply(subgoal_tac " dom  ((\<lambda>la. subst_option (Suc n) (lift s' 0) (t'a la))(l \<mapsto> the (t l)[lift s' 0/Suc n])) = dom t")
apply(simp)
apply(force)
apply(simp)
apply(simp add: dom_substoption_lem)
apply(force)
apply(simp)
apply(rule ext)
apply(case_tac "l=la")
apply(force)
apply(subgoal_tac "the (t la) = the (t'a la)")
apply(simp)
apply(rule conjI)
apply(rule impI)
apply(simp)
apply(rule impI)

apply(force)
apply(rotate_tac -3)
apply(erule ssubst)
apply(subgoal_tac " ((t'a(l \<mapsto> the (t l))) la) =  (t'a la)")
apply(rotate_tac -1)
apply(erule ssubst)
apply(rule refl)
apply(rule  insert_select2)
apply(assumption)
apply(simp)
apply(rule ext)+
apply(case_tac "l=la")
apply(simp)
apply(simp)
done



lemma par_beta_subst [rule_format]:
    "\<forall>s s' t' n. s => s' --> t => t' --> t[s/n] => t'[s'/n]";
apply (simp add: par_beta_subst_prep);
done;


lemma one_more_dom :"!! f .  \<lbrakk>f \<noteq> empty\<rbrakk> \<Longrightarrow> \<forall> l \<in> dom f . \<exists> f' . f = f'(l \<mapsto> the(f l) ) \<and> l \<notin> dom f' "
apply(rule ballI)
apply(rule_tac x="\<lambda> la . if la = l then None else f la" in exI )
apply(rule conjI)
apply(rule ext)
apply(case_tac "x=l")
apply(auto)

done


lemma fmap_inducta: " \<lbrakk>P empty;
                !! x (F::Label -~>dB) y . \<lbrakk>  P F; x \<notin> dom F \<rbrakk>
                \<Longrightarrow> P (F(x \<mapsto> y))  \<rbrakk> \<Longrightarrow> P F";
thm fmap_induct
thm list_induct3

apply(rule_tac  fmap_induct)
apply(simp)
apply(blast)
done

lemma fmap_induct3: "!!   (F2::Label -~>dB)  (F3::Label -~>dB) . \<lbrakk> dom  (F1::Label -~>dB) = dom F2; dom F3 = dom F1; P empty empty empty;
                !! x a b c (F1::Label -~>dB)  (F2::Label -~>dB)  (F3::Label -~>dB)  . \<lbrakk>  P F1 F2 F3; dom F1 = dom F2; dom F3 = dom F1;  x \<notin> dom F1 \<rbrakk>
                \<Longrightarrow> P (F1(x \<mapsto> a)) (F2(x \<mapsto> b)) (F3(x \<mapsto> c))  \<rbrakk> \<Longrightarrow> P F1 F2 F3";
thm fmap_inducta

thm list.induct
apply(induct F1 rule: fmap_inducta)
defer
apply(case_tac "F2=empty");
defer





apply(case_tac "F3=empty");
defer

apply(frule one_more_dom)
apply(erule_tac x=x in ballE)
apply(subgoal_tac "F2 \<noteq>  empty")
apply(frule_tac f=F3 in  one_more_dom)
apply(subgoal_tac "x \<in> dom F2")
apply(subgoal_tac "x \<in> dom F3")
thm ballE
apply(rotate_tac -3)


apply(erule_tac x=x in ballE)
apply(erule exE)+
apply(erule conjE)+
apply(subgoal_tac "dom F = dom f'")
apply(subgoal_tac "dom f'a = dom F")
apply(rotate_tac -4)
apply(erule ssubst)
apply(rotate_tac -2)
apply(erule ssubst)

apply(subgoal_tac "P F f' f'a")


apply(subgoal_tac "dom f'a = dom F")
apply(drule_tac x="f'"  in meta_spec)

apply(rotate_tac -2)
apply(drule_tac x="f'a"  in meta_spec)
apply(drule_tac x="x"  in meta_spec)
apply(drule_tac x="y"  in meta_spec)
apply(drule_tac x="the (F2 x)"  in meta_spec)
apply(drule_tac x="the (F3 x)"  in meta_spec)
apply(drule_tac x="F"  in meta_spec)
apply(drule_tac x="f'"  in meta_spec)
apply(drule_tac x="f'a"  in meta_spec)
apply(blast)

apply(force)
apply(force)
apply(rule insert_dom_less_eq)
apply(simp )
apply(assumption)

apply(subgoal_tac "! y. (dom (f'a(x \<mapsto> y))) = dom F3")
apply(simp)
apply(rule allI)
apply(force)
apply(rule insert_dom_less_eq)
apply(simp)
apply(assumption)
apply(subgoal_tac "! y. (dom (f'(x \<mapsto> y))) = dom F2")
apply(simp)
apply(rule allI)
apply(force)
apply(force)
apply(force)
apply(force)
apply(assumption)
apply(frule one_more_dom)
apply(erule_tac x=x in ballE)
apply(subgoal_tac "F2 \<noteq>  empty")
apply(frule_tac f=F3 in  one_more_dom)
apply(subgoal_tac "x \<in> dom F2")
apply(subgoal_tac "x \<in> dom F3")
thm ballE
apply(rotate_tac -3)


apply(erule_tac x=x in ballE)
apply(erule exE)+
apply(erule conjE)+
apply(subgoal_tac "dom F = dom f'")
apply(subgoal_tac "dom f'a = dom F")
apply(rotate_tac -4)
apply(erule ssubst)
apply(rotate_tac -2)
apply(erule ssubst)

apply(subgoal_tac "P F f' f'a")


apply(subgoal_tac "dom f'a = dom F")
apply(drule_tac x="f'"  in meta_spec)

apply(rotate_tac -2)
apply(drule_tac x="f'a"  in meta_spec)
apply(drule_tac x="x"  in meta_spec)
apply(drule_tac x="y"  in meta_spec)
apply(drule_tac x="the (F2 x)"  in meta_spec)
apply(drule_tac x="the (F3 x)"  in meta_spec)
apply(drule_tac x="F"  in meta_spec)
apply(drule_tac x="f'"  in meta_spec)
apply(drule_tac x="f'a"  in meta_spec)
apply(blast)

apply(force)
apply(force)
apply(rule insert_dom_less_eq)
apply(simp )
apply(assumption)

apply(subgoal_tac "! y. (dom (f'a(x \<mapsto> y))) = dom F3")
apply(simp)
apply(rule allI)
apply(force)
apply(rule insert_dom_less_eq)
apply(simp)
apply(assumption)
apply(subgoal_tac "! y. (dom (f'(x \<mapsto> y))) = dom F2")
apply(simp)
apply(rule allI)
apply(force)
apply(force)
apply(force)
apply(force)
apply(assumption)
apply(subgoal_tac "x \<in>  dom F2")
apply(simp)
apply(subgoal_tac "dom F2 = dom (F(x \<mapsto> y)) ")
apply(simp)
apply(force)
apply(subgoal_tac "F2 = empty")
apply(subgoal_tac "F3 = empty")
apply(simp)
apply(simp add:  empty_dom)

apply(simp add:  empty_dom)
apply(force)
apply(force)
done




(* Very technical transformation of a !? into a ?! for lists and par_red *)
lemma ex_list_ex0: " \<lbrakk> dom  (s::Label -~>dB)  = dom  (s'::Label -~>dB) ; dom  (lz::Label -~>dB)  = dom s \<rbrakk> \<Longrightarrow>
             ( \<forall> l \<in> dom s . the (s l) => the (lz l))
           & (\<forall>  l \<in> dom s. (! z. the (s  l) => z \<longrightarrow> (? u. the (s'  l) => u \<and> z => u)))
          \<longrightarrow> (\<exists>  lu. (dom lu = dom s) & (\<forall>  l \<in>  dom  s. the (s'  l) => the (lu  l) \<and> the (lz  l) => the (lu  l)))";
apply(erule fmap_induct3)
apply(subst refl)
apply(assumption)
apply(simp)
apply(rule_tac x="empty" in exI)
apply(simp add: empty_dom)
apply(rule impI)
apply(erule impE)
apply(rule conjI)
apply(rule ballI)
apply(erule conjE)
apply(erule_tac x=l in ballE)
thm par_beta_less'
apply(subgoal_tac "x \<notin>  dom F3")
apply(case_tac "l=x")
apply(simp)
apply(simp)
apply(simp)
apply(force) (*contradiction*)
apply(rule ballI)+
apply(rule allI)
apply(rule impI)
apply(erule conjE)
apply(rotate_tac -1)
apply(erule_tac x=l in ballE)
apply(erule_tac x=z in allE)
apply(erule impE)
apply(subgoal_tac "the (F1 l) =  the ((F1(x \<mapsto> a)) l)")
apply(simp)
apply(case_tac "x=l")
apply(simp)
apply(subst insert_select2)
apply(assumption)
apply(rule refl)
apply(erule exE)
apply(rule_tac x=u in exI) 
apply(erule conjE)
apply(rule conjI)
apply(subgoal_tac "the (F2 l) =  the ((F2(x \<mapsto> b)) l)")
apply(simp)
apply(case_tac "x=l")
apply(simp)
apply(subst insert_select2)
apply(simp)
apply(rule refl)
apply(assumption)
apply(force) (*contradiction*)

apply(erule exE)
apply(erule conjE)+
apply(rotate_tac 5)
apply(erule_tac x=x in ballE)
apply(erule_tac x=" the ((F3(x \<mapsto> c)) x)" in allE)
apply(erule impE)
apply(rotate_tac -1)
apply(erule_tac x=x in ballE)
apply(assumption)
apply(force) (*contradiction*)
prefer 2
apply(force) (*contradiction*)
apply(erule exE)
apply(erule conjE)

apply(rule_tac x="lu(x |-> u)" in exI)
apply(rule conjI)+


apply(erule insert_dom_eq)
apply(rule ballI)
apply(rule conjI)

apply(case_tac "l=x")
apply(simp)
apply(subgoal_tac "l \<in> dom F1") 
apply(rotate_tac -3)
apply(erule_tac x=l in ballE)
apply(erule conjE)
apply(simp)
apply(force)
apply(force)
apply(case_tac "l=x")
apply(simp)
apply(erule_tac x=l in ballE)
apply(simp)
apply(force) (*contradiction*)
done



lemma ex_list_ex: "\<lbrakk> dom (s::Label -~>dB) = dom (s'::Label -~>dB); dom s = dom (lz::Label -~>dB); \<forall>  l \<in> dom s. the (s l) => the (lz l);
           (\<forall> l \<in> dom s. (\<forall>z. the (s l) => z \<longrightarrow> (\<exists>u. the (s' l) => u \<and> z => u))) \<rbrakk>
           \<Longrightarrow> (\<exists>  lu. dom lu = dom s \<and>  (\<forall>  l \<in> dom s. the (s' l) => the (lu  l) \<and> the (lz  l) => the (lu  l)))";
apply (drule ex_list_ex0);
apply (erule sym);
apply (erule mp);
apply (erule conjI);
by assumption;



subsection {* Confluence (directly) *}

(*** Main result: Confluence of beta relation for Sigma calculus              ***)
(*** by diamond property of parallel reduction and beta <= par_beta <= beta^* ***)


lemma diamond_par_beta: "diamond par_beta"
  apply (unfold diamond_def commute_def square_def)
  apply (rule impI [THEN allI [THEN allI]])
  apply (erule par_beta.induct)
apply simp;
(* Obj s *)
apply (rule allI);
apply (rule impI);
apply (frule Obj_par_red);
apply (erule exE);
apply (erule conjE);
apply (rotate_tac -1);
apply (subgoal_tac "Obj s => Obj lz");
apply (erule ssubst);
apply (subgoal_tac "? lu. (dom lu = dom s' & (\<forall>  l \<in> dom s'. the (s'  l) => the (lu  l))) &
                          (dom lu = dom lz & (\<forall>  l \<in> dom lz. the (lz  l) => the (lu  l)))");
apply (erule exE);
apply (erule conjE)+;
apply (rule_tac x = "Obj lu" in exI);
apply (simp add: rtrancl_beta_obj_n);

apply (frule ex_list_ex);
apply (rotate_tac -2);
apply assumption;
apply(rule obj_par_beta)
apply assumption;
apply simp;
apply (erule exE);
apply (rule_tac x = lu in exI);
apply simp;
apply simp;
(* *)

apply (rule allI);
apply (rule impI);

apply (frule Upd_par_red);
apply (erule disjE);
apply (erule exE);
apply (erule exE);
apply (erule conjE)+;
apply (subgoal_tac "Upd s l t => Upd s'a l t'a");
apply (rotate_tac -2);
apply (erule ssubst);
apply(rotate_tac -4)
apply (drule_tac x = "t'a" in spec);
apply (erule impE);
apply assumption;
apply (drule_tac x = "s'a" in spec);
apply (erule impE);
apply assumption;
apply (erule exE);
apply (erule exE);
apply (rule_tac x = "Upd ua l u" in exI);
apply (erule conjE)+;
apply (rule conjI);
apply (erule upd);
apply assumption;
apply (erule upd);
apply assumption;
apply simp;
(* upd: case Obj *)
apply (erule exE)+;
apply (erule conjE)+;
apply (subgoal_tac "Upd s l t => Obj (f'(l |-> t'a))");
apply (rotate_tac -2);
apply (erule ssubst);
apply simp;
apply (rotate_tac 2 );
apply (drule_tac x = "Obj f'" in spec);
apply (erule impE);
apply assumption;
apply (drule_tac x = "t'a" in spec);
apply (erule impE);
apply assumption;
apply (erule exE);
apply (erule exE);
apply (erule conjE)+;
apply (rotate_tac -3);
apply (frule Obj_par_red);
apply (erule exE);
apply (erule conjE);

apply (rule_tac x = "Obj (lz(l |-> ua))" in exI);
apply (rule conjI);
apply (rotate_tac 1);
apply (frule Obj_par_red);
apply (erule exE);
apply (erule conjE);
apply (rotate_tac -1);
thm upd'
apply(subgoal_tac "\<exists> f'' . s' = Obj f''")
apply(erule exE)
apply (rotate_tac -1);
apply(subgoal_tac "s' = Obj f''")
apply(erule ssubst)
apply(rule upd')
apply(force)
apply(assumption)
prefer 3
apply(force)
apply(subgoal_tac "dom f'' = dom f")
apply(case_tac "l \<in> dom f")
apply(simp)
apply(force)
apply(force)
apply(assumption)
apply(force)


(* Upd Obj  *)

apply(rotate_tac -4)
apply(subgoal_tac "s = Obj f")
apply(erule ssubst)
apply(rule upd')
apply(assumption)
apply(assumption)
apply(case_tac "l \<in> dom f")
apply(simp)
apply(force)
apply(assumption)

apply (rule allI);
apply (rule impI);
apply (rotate_tac -1);
apply (frule Upd_par_red);
apply (erule disjE);
apply (erule exE)+;
apply (erule conjE)+;
apply (rotate_tac -3);
apply (frule Obj_par_red);
apply (erule exE);
apply (erule conjE);
apply (drule_tac x = "Obj lz" in spec);
apply (erule impE);
apply simp;
apply (drule_tac x = "t'a" in spec);
apply (erule impE);
apply assumption;
apply (erule exE);
apply (erule exE);
apply (erule conjE)+;
apply (rotate_tac -4);
apply (frule Obj_par_red);
apply (frule Obj_par_red);
apply (erule exE)+;
apply (erule conjE)+;
apply (rule_tac x = "Obj (lza (l |-> ua))" in exI);
apply (rule conjI);
apply (rule par_beta_obj_subst);
apply assumption;
apply simp;
apply simp;
(* *)
apply (erule exE)+;
apply (erule conjE)+
apply (rotate_tac -1);
apply simp;
apply (drule_tac x = "Obj f'" in spec);
apply (erule impE);
apply assumption;
apply (erule exE);
apply (erule conjE);
apply (drule_tac x = "t'a" in spec);
apply (erule impE);
apply assumption;
apply (erule exE);
apply (erule conjE);
apply (rotate_tac 6);
apply (frule Obj_par_red);
apply (frule Obj_par_red);
apply (erule exE)+;
apply (erule conjE)+;
apply (rule_tac x = "Obj (lza (l |-> ua))" in exI);
apply (simp add: par_beta_obj_subst);
(* Call *)
apply (rule allI);
apply (rule impI);
apply (frule Call_par_red);
apply (erule disjE);
apply (erule exE);
apply (erule conjE);
apply (drule_tac x = ta in spec);
apply (erule impE);
apply assumption;
apply (erule exE);
apply (erule conjE);
apply (rule_tac x = "Call u l" in exI);
apply simp;
(* *)
defer;
apply (rule allI);
apply (rule impI);
apply (frule Call_par_red);
apply (erule disjE)
apply (erule exE);
apply (erule conjE);
apply (rotate_tac -2);
apply (frule Obj_par_red);
apply (erule exE);
apply (erule conjE);
apply (drule_tac x = "Obj lz" in spec);
apply (erule impE);
apply simp;
apply (erule exE);
apply (erule conjE);
apply (rotate_tac -2);
apply (frule Obj_par_red);
apply (frule Obj_par_red);
apply (erule exE)+;
apply (erule conjE)+;
apply (rule_tac x = "the (lza  l) [Obj lza / 0]" in exI);
apply (rule conjI);
apply (rule par_beta_subst);
apply(simp)
(* make this a lemma afterwards - it's conceptually important *)
apply (subgoal_tac "! f g. Obj f => Obj g & l \<in> dom f  \<longrightarrow> the (f  l) => the (g  l)"); 
apply simp;
apply(rule allI)
apply(rule allI)
apply(rule impI)apply(erule conjE)
apply(simp)
thm obj_par_beta
apply(frule_tac f=fa and f'=g in  obj_par_beta)
apply(erule_tac x=l in ballE)
apply(assumption)
apply(simp)



apply (erule ssubst);
apply (rotate_tac -5 );
apply (erule ssubst);
apply (rule beta);
apply simp;
apply(case_tac "l \<in> dom f")
apply(simp)
apply(simp)

(* apply (rule beta); *)

(* snd case of Call *)
apply (erule exE)+;
apply (erule conjE)+;
apply (drule_tac x = "Obj f'a" in spec);
apply (erule impE);
apply simp;
apply (erule exE);
apply (erule conjE);
apply (rotate_tac -2);
apply (frule Obj_par_red);
apply (frule Obj_par_red);
apply (erule exE)+;
apply (erule conjE)+;
apply (rule_tac x = "the (lz  l)[Obj lz / 0]" in exI);
apply (rule conjI);
apply (rule par_beta_subst);
(* This lemma again s.o. *)
apply (subgoal_tac "! f g. Obj f => Obj g & l \<in> dom f  \<longrightarrow> the (f  l) => the (g  l)"); 
apply simp;
apply(rule allI)
apply(rule allI)
apply(rule impI)apply(erule conjE)
apply(simp)
thm obj_par_beta
apply(frule_tac f=fb and f'=g in  obj_par_beta)
apply(erule_tac x=l in ballE)
apply(assumption)
apply fastsimp;
apply(subgoal_tac "Obj f' => Obj lz")
apply(frule_tac f=f' and f'=lz in obj_par_beta)
apply(erule_tac x=l in ballE)
apply(assumption)
apply(simp)
apply(simp)
apply (rotate_tac 8);
apply (erule ssubst);
apply (rule par_beta_subst);
(* This lemma again s.o. *)
apply (subgoal_tac "! f g. Obj f => Obj g & l \<in> dom f  \<longrightarrow> the (f l) => the (g l)"); 
apply(simp)
apply(rule allI)
apply(rule allI)
apply(rule impI)apply(erule conjE)
apply(simp)
thm obj_par_beta
apply(frule_tac f=fb and f'=g in  obj_par_beta)
apply(erule_tac x=l in ballE)
apply(assumption)
apply simp;
apply(subgoal_tac "Obj f'a => Obj lz")
apply(frule_tac f=f'a and f'=lz in obj_par_beta)
apply(erule_tac x=l in ballE)
apply fastsimp;
apply fastsimp;
apply(simp)


(* i am here *)



apply(erule exE)+
apply(erule conjE)
apply(erule conjE)
apply(erule conjE)
apply(erule_tac x="Obj f'" in allE)
apply(erule impE)
apply(simp)
apply(erule exE)
apply(erule conjE)
apply(subgoal_tac "\<exists> f'' . u = Obj f''") 
prefer 2
apply(force)
apply(erule exE)
apply(subgoal_tac "Obj f => t")
prefer 2
apply(force)
apply(subgoal_tac "\<exists> f''' . t = Obj f'''") 
prefer 2
apply(force)
apply(erule exE)
apply(rule_tac x="the (f'' l)[Obj f''/0]" in exI)
apply(rule conjI)
apply(subgoal_tac "t = Obj f'''") 
apply(rotate_tac -1)
apply(erule ssubst)
apply(rule beta)
apply(simp)
apply(simp)
apply(subgoal_tac "dom f''' = dom f''")
apply(subgoal_tac "dom f' = dom f''")

apply(rotate_tac -1)

apply(erule subst)
apply(assumption)
apply(rule lift_obj_lem00)
apply(assumption)
apply(rule lift_obj_lem00)
apply(assumption)
apply(assumption)
apply(subgoal_tac " the (f' l) => the (f'' l) ")
apply(rotate_tac 5)
apply(erule ssubst)
apply(rule par_beta_subst)
apply(simp)

apply(frule obj_par_beta)
apply(erule_tac x=l in ballE)
apply(assumption)
apply(simp)
apply(subgoal_tac "Obj f' => Obj f''")
apply(frule_tac f=f' and f'=f'' in obj_par_beta)
apply(erule_tac x=l in ballE)
apply(assumption)
apply(simp)
apply(simp)
done




subsection {* Confluence (classical not via complete developments) *}

theorem beta_confluent: "confluent beta"
  apply (rule diamond_par_beta diamond_to_confluence
    par_beta_subset_beta beta_subset_par_beta)+
  done



end

